"""
Reading helpers.
"""

import functools
from lute.models.term import Term, Status
from lute.models.book import Text
from lute.book.stats import mark_stale
from lute.read.render.service import get_paragraphs, find_all_Terms_in_string
from lute.read.render.renderable_calculator import TokenLocator
from lute.term.model import Repository
from lute.db import db

# from lute.utils.debug_helpers import DebugTimer


def set_unknowns_to_known(text: Text):
    """
    Given a text, create new Terms with status Well-Known
    for any new Terms.
    """
    language = text.book.language
    paragraphs = get_paragraphs(text.text, text.book.language)
    # Just in case.
    _add_status_0_terms(paragraphs, language)

    tis = [
        ti
        for para in paragraphs
        for sentence in para
        for ti in sentence.textitems
        if ti.is_word
    ]

    def is_unknown(ti):
        return ti.is_word == 1 and ti.term is not None and ti.wo_status == 0

    unknown_ids = [t.wo_id for t in tis if is_unknown(t)]
    uniques = list(set(unknown_ids))

    batch_size = 100
    i = 0

    for u in uniques:
        t = Term.find(u)
        t.status = Status.WELLKNOWN
        db.session.add(t)
        i += 1
        if i % batch_size == 0:
            db.session.commit()

    # Commit any remaining.
    db.session.commit()


def bulk_status_update(text: Text, terms_text_array, new_status):
    """
    Given a text and list of terms, update or create new terms
    and set the status.
    """
    language = text.book.language
    repo = Repository(db)
    for term_text in terms_text_array:
        t = repo.find_or_new(language.id, term_text)
        t.status = new_status
        repo.add(t)
    repo.commit()


def _create_unknown_terms(textitems, lang):
    "Create any terms required for the page."
    # dt = DebugTimer("create-unk-terms")
    toks = [t.text for t in textitems]
    # print(f"creating toks {toks}", flush=True)
    unique_word_tokens = list(set(toks))
    # print(f"creating unique toks {unique_word_tokens}", flush=True)
    all_new_terms = [Term.create_term_no_parsing(lang, t) for t in unique_word_tokens]
    # print(f"all_new_terms = {all_new_terms}", flush=True)
    # dt.step("make all_new_terms")

    unique_text_lcs = {}
    for t in all_new_terms:
        if t.text_lc not in unique_text_lcs:
            unique_text_lcs[t.text_lc] = t
    unique_new_terms = unique_text_lcs.values()
    # print(f"utlcs keys = {unique_text_lcs.keys()}", flush=True)
    # dt.step("find unique_new_terms")

    for t in unique_new_terms:
        t.status = 0
        db.session.add(t)
    db.session.commit()
    # dt.step("commit")
    # dt.summary()

    return unique_new_terms


def _add_status_0_terms(paragraphs, lang):
    "Add status 0 terms for new textitems in paragraph."
    new_textitems = [
        ti
        for para in paragraphs
        for sentence in para
        for ti in sentence.textitems
        if ti.is_word and ti.term is None
    ]
    # Create new terms for all unknown word tokens in the text.
    new_terms = _create_unknown_terms(new_textitems, lang)

    # Set the terms for the unknown_textitems
    textlc_to_term_map = {}
    for t in new_terms:
        textlc_to_term_map[t.text_lc] = t
    # print("map: textlc_to_term_map")
    # for k, v in textlc_to_term_map.items():
    #     print(f"{k}: {v}", flush=True)
    for ti in new_textitems:
        # print(f'Assigning term from map to ti with ti.text_lc = "{ti.text_lc}"')
        ti.term = textlc_to_term_map[ti.text_lc]


def start_reading(dbbook, pagenum, db_session):
    "Start reading a page in the book, getting paragraphs."

    text = dbbook.text_at_page(pagenum)
    text.load_sentences()

    mark_stale(dbbook)
    dbbook.current_tx_id = text.id
    db_session.add(dbbook)
    db_session.add(text)
    db_session.commit()

    lang = text.book.language
    paragraphs = get_paragraphs(text.text, lang)
    _add_status_0_terms(paragraphs, lang)

    return paragraphs


def get_popup_data(termid):
    "Get the data necessary to render a term popup."
    term = Term.find(termid)

    term_tags = [tt.text for tt in term.term_tags]

    def make_array(t):
        ret = {
            "term": t.text,
            "roman": t.romanization,
            "trans": t.translation if t.translation else "-",
            "tags": [tt.text for tt in t.term_tags],
        }
        return ret

    parent_terms = [p.text for p in term.parents]
    parent_terms = ", ".join(parent_terms)

    parents = term.parents
    if len(parents) == 1 and parents[0].translation == term.translation:
        parents = []
    parent_data = [make_array(p) for p in parents]

    def sort_components(components):
        # Sort components by min position in string and length.
        subj = TokenLocator.make_string(term.text)
        tocloc = TokenLocator(term.language, subj)
        component_and_pos = []
        for c in components:
            locs = tocloc.locate_string(c.text)
            # pylint: disable=consider-using-generator
            index = min([loc["index"] for loc in locs])
            component_and_pos.append([c, index])

        def compare(a, b):
            # Lowest position (closest to front of string) sorts first.
            if a[1] != b[1]:
                return -1 if (a[1] < b[1]) else 1
            # Longest sorts first.
            alen = len(a[0].text)
            blen = len(b[0].text)
            return -1 if (alen > blen) else 1

        component_and_pos.sort(key=functools.cmp_to_key(compare))
        return [c[0] for c in component_and_pos]

    components = [
        c for c in find_all_Terms_in_string(term.text, term.language) if c.id != term.id
    ]
    components = sort_components(components)

    component_data = [make_array(c) for c in components]
    component_data = [c for c in component_data if c["trans"] != "-"]

    images = [term.get_current_image()] if term.get_current_image() else []
    for p in term.parents:
        if p.get_current_image():
            images.append(p.get_current_image())
    # DISABLED CODE: Don't include component images in the hover for now,
    # it can get confusing!
    # ref https://github.com/LuteOrg/lute-v3/issues/355
    # for c in components:
    #     if c.get_current_image():
    #         images.append(c.get_current_image())

    images = list(set(images))

    return {
        "term": term,
        "flashmsg": term.get_flash_message(),
        "term_tags": term_tags,
        "term_images": images,
        "parentdata": parent_data,
        "parentterms": parent_terms,
        "components": component_data,
    }



"""
Backup routes.

Backup settings form management, and running backups.
"""

import os
import traceback
from flask import (
    Blueprint,
    current_app,
    render_template,
    request,
    jsonify,
    redirect,
    send_file,
    flash,
)
from lute.models.setting import BackupSettings
from lute.backup.service import create_backup, skip_this_backup, list_backups


bp = Blueprint("backup", __name__, url_prefix="/backup")


@bp.route("/index")
def index():
    """
    List all backups.
    """
    settings = BackupSettings.get_backup_settings()
    backups = list_backups(settings.backup_dir)
    backups.sort(reverse=True)

    return render_template(
        "backup/index.html", backup_dir=settings.backup_dir, backups=backups
    )


@bp.route("/download/<filename>")
def download_backup(filename):
    "Download the given backup file."
    settings = BackupSettings.get_backup_settings()
    fullpath = os.path.join(settings.backup_dir, filename)
    return send_file(fullpath, as_attachment=True)


@bp.route("/backup", methods=["GET"])
def backup():
    """
    Endpoint called from front page.

    With extra arg 'type' for manual.
    """
    backuptype = "automatic"
    if "type" in request.args:
        backuptype = "manual"

    settings = BackupSettings.get_backup_settings()
    return render_template(
        "backup/backup.html", backup_folder=settings.backup_dir, backuptype=backuptype
    )


@bp.route("/do_backup", methods=["POST"])
def do_backup():
    """
    Ajax endpoint called from backup.html.
    """
    backuptype = "automatic"
    prms = request.form.to_dict()
    if "type" in prms:
        backuptype = prms["type"]

    c = current_app.env_config
    settings = BackupSettings.get_backup_settings()
    is_manual = backuptype.lower() == "manual"
    try:
        f = create_backup(c, settings, is_manual=is_manual)
        flash(f"Backup created: {f}", "notice")
        return jsonify(f)
    except Exception as e:  # pylint: disable=broad-exception-caught
        tb = traceback.format_exc()
        return jsonify({"errmsg": str(e) + " -- " + tb}), 500


@bp.route("/skip_this_backup", methods=["GET"])
def handle_skip_this_backup():
    "Update last backup date so backup not attempted again."
    skip_this_backup()
    return redirect("/", 302)



"""
Show terms in datatables.
"""

from lute.db import db
from lute.utils.data_tables import DataTablesSqliteQuery


def get_data_tables_list(parameters):
    "json data for datatables."
    base_sql = """SELECT
          TgID,
          TgText,
          TgComment,
          ifnull(TermCount, 0) as TermCount
          FROM tags
          left join (
            select WtTgID,
            count(*) as TermCount
            from wordtags
            group by WtTgID
          ) src on src.WtTgID = TgID
    """
    session = db.session
    connection = session.connection()
    return DataTablesSqliteQuery.get_data(base_sql, parameters, connection)



"""
Parsing using Jieba

The parser uses jieba to do parsing and pypinyin for character readings

Includes classes:

- MandarinParser

"""

import re
from typing import List
import jieba
from pypinyin import pinyin
from lute.parse.base import ParsedToken, AbstractParser


class MandarinParser(AbstractParser):
    """
    A parser for Mandarin Chinese,
    using the jieba library for text segmentation.
    """

    @classmethod
    def name(cls):
        return "Lute Mandarin Chinese"

    def get_parsed_tokens(self, text: str, language) -> List[ParsedToken]:
        """
        Returns ParsedToken array for given language.
        """

        # Ensure standard carriage returns so that paragraph
        # markers are used correctly.  Lute uses paragraph markers
        # for rendering.
        text = text.replace("\r\n", "\n")

        words = list(jieba.cut(text))
        tokens = []
        pattern = f"[{language.word_characters}]"
        for word in words:
            is_word_char = re.match(pattern, word) is not None
            is_end_of_sentence = word in language.regexp_split_sentences
            if word == "\n":
                word = "¶"
            if word == "¶":
                is_word_char = False
                is_end_of_sentence = True
            p = ParsedToken(word, is_word_char, is_end_of_sentence)
            tokens.append(p)
        return tokens

    def get_reading(self, text: str):
        """
        Get the pinyin for the given text.
        Returns None if the text is all Chinese characters, or the pinyin
        doesn't add value (same as text).
        """
        # Use pypinyin to get the pinyin of the text
        pinyin_list = pinyin(text)
        # Flatten the list of lists to a single list
        pinyin_list = (item for sublist in pinyin_list for item in sublist)
        # Join the pinyin into a single string
        ret = " ".join(pinyin_list)
        if ret in ("", text):
            return None
        return ret



"""
Generate a rough data file for a given language for all terms.

Gets *all* books for a given language, and writes a data file to
the specified csv output file name.

Generates csv with headings:
term; count; familycount; books; definition; status; children

e.g.
term; count; familycount; books; definition; status; children
haber; 100; 1500; book1,book2; to exist; 99; hay (500), he (200), has (150) ...

There is probably a far better way to do this, likely using something
fairly heavyweight like pandas.  This works well enough for now.
"""

import sys
import csv
from collections import Counter
from lute.db import db
from lute.models.book import Book
from lute.term.model import Repository


def get_dist(book, collector, termrepo, language_id):  # pylint: disable=too-many-locals
    """
    Get word distribution in book.

    The data is added to the collector dictionary.
    """

    # Get all terms and counts.
    fulltext = "\n".join([t.text for t in book.texts])
    pts = book.language.get_parsed_tokens(fulltext)
    words = [pt.token for pt in pts if pt.is_word]

    # distrib = { 'term1': count1, 'term2': count2, ... }
    distrib = dict(Counter(words))

    # The distribution doesn't handle capitalization, so it will
    # contain things like { 'There': 10, 'there': 20 }.  Do a lookup
    # for each term ('There', 'there') in the repository to see if a
    # matching term with a standardized term.text is found.
    normalized = {}

    totcount = len(distrib.keys())
    i = 0
    print(f"Loading data for book {book.title} ...")
    for k, v in distrib.items():
        i += 1
        if i % 100 == 0:
            print(f"  {i} of {totcount}", end="\r")
        norm_word = termrepo.find_or_new(language_id, k)
        norm_entry = normalized.get(norm_word.text, {"count": 0, "parents": []})
        norm_entry["count"] += v
        norm_entry["parents"] = norm_word.parents
        normalized[norm_word.text] = norm_entry

    # normalized = { 'there': { 'count': 30, 'parents': [...] }, ... }.
    #
    # The collector may already have the term ('there') from prior
    # books, so combine those.
    for t, n in normalized.items():  # pylint: disable=redefined-outer-name
        entry = collector.get(t, {"term": t, "count": 0, "books": []})
        entry["count"] += n["count"]
        entry["books"].append(book.title)
        collector[t] = entry

        # The term may have a parent that isn't actually present in any book!
        # We need to add those parents to the collector as well, or later
        # searches for the parent will fail.
        for p in n["parents"]:
            pentry = collector.get(p, {"term": p, "count": 0, "books": []})
            collector[p] = pentry


def _load_hash_from_term(t, term):
    "Load common data to hash."
    t["parent"] = ", ".join(term.parents)
    t["definition"] = term.translation or "-"
    t["status"] = term.status if term.id is not None else "-"
    t["children"] = "-"
    t["childbooks"] = []
    t["tags"] = ", ".join(term.term_tags)


def load_term_data(langid, terms, repo):
    "Load basic data."
    totcount = len(terms.keys())
    i = 0
    print("Loading term data ...")
    for k, t in terms.items():  # pylint: disable=unused-variable
        i += 1
        if i % 100 == 0:
            print(f"  {i} of {totcount}", end="\r")

        term = repo.find_or_new(langid, t["term"])
        _load_hash_from_term(t, term)
        t["familycount"] = t["count"]


def load_parent_data(langid, terms, repo):
    "Get and print data."

    parents = list({t["parent"] for t in terms.values() if t["parent"] != ""})

    missingparents = [p for p in parents if p not in terms]
    totcount = len(missingparents)
    i = 0
    print("Loading missing parents ...")
    for p in missingparents:
        i += 1
        if i % 100 == 0:
            print(f"  {i} of {totcount}", end="\r")

        term = repo.find_or_new(langid, p)
        t = {"term": p, "count": 0, "books": []}
        _load_hash_from_term(t, term)
        t["familycount"] = 0
        terms[p] = t

    totcount = len(parents)
    i = 0
    print("Finalizing parent data ...")
    for p in parents:
        i += 1
        if i % 100 == 0:
            print(f"  {i} of {totcount}", end="\r")

        children = [c for c in terms.values() if c["parent"] == p]
        csorted = sorted(children, key=lambda c: c["count"], reverse=True)
        children_string = "; ".join([f"{c['term']} ({c['count']})" for c in csorted])
        childbooks = [c["books"] for c in children]
        childbooks = list({b for blist in childbooks for b in blist})
        childtotcount = sum(c["count"] for c in children)

        terms[p]["children"] = children_string
        terms[p]["childbooks"] = childbooks
        terms[p]["familycount"] += childtotcount


def get_output_data(terms):
    "Get the final set of output data."
    printterms = [
        t for t in terms.values() if t["parent"] == "" or t["children"] != "-"
    ]

    # Clean up data for printing.
    for t in printterms:
        t["books"] = list(set(t["books"] + t["childbooks"]))
        t["books"] = "; ".join(t["books"])
        del t["childbooks"]

    return printterms


def generate_file(language_name, outfile_name):
    """
    Generate the datafile for the language.
    """
    books = db.session.query(Book).all()
    books = [b for b in books if b.language.name == language_name]
    if len(books) == 0:
        print(f"No books for given language {language_name}, quitting.")
        sys.exit(0)

    langid = books[0].language.id

    repo = Repository(db)
    terms = {}
    for b in books:
        get_dist(b, terms, repo, langid)

    load_term_data(langid, terms, repo)
    load_parent_data(langid, terms, repo)
    outdata = get_output_data(terms)

    ptsorted = sorted(outdata, key=lambda c: c["familycount"], reverse=True)
    keys = [
        "term",
        "count",
        "familycount",
        "books",
        "definition",
        "status",
        "children",
        "tags",
    ]
    print(f"Writing to {outfile_name}")
    with open(outfile_name, "w", newline="", encoding="utf-8") as outfile:
        writer = csv.DictWriter(outfile, fieldnames=keys, extrasaction="ignore")
        writer.writeheader()
        for r in ptsorted:
            writer.writerow(r)
    print("Done.")



"""
Test main entry point.
"""

import os
import sqlite3
from contextlib import closing

from lute.config.app_config import AppConfig
from lute.app_factory import create_app


def test_init_no_existing_database(testconfig):
    """
    If no db exists, init should:
    - do the db setup
    - create the flask app
    """
    if os.path.exists(testconfig.dbfilename):
        os.unlink(testconfig.dbfilename)

    config_file = AppConfig.default_config_filename()
    app = create_app(config_file)

    assert os.path.exists(testconfig.dbfilename) is True, "db exists"
    assert testconfig.dbname.startswith("test_")

    def assert_tables_exist(msg: str):
        with closing(sqlite3.connect(testconfig.dbfilename)) as conn:
            cur = conn.cursor()
            sql = """SELECT name FROM sqlite_master
            WHERE type='table' AND name in ('words', 'languages')
            order by name;"""
            tnames = cur.execute(sql).fetchall()
            tnames = [t[0] for t in tnames]
            assert tnames == ["languages", "words"], msg

    assert_tables_exist("db set up correctly")

    client = app.test_client()
    response = client.get("/")
    assert b"Lute" in response.data



"""
Lute settings, in settings key-value table.
"""

import os
import datetime
import time
from flask import current_app
from lute.db import db


class SettingBase(db.Model):
    """
    Main settings table.

    This class should not be used, it's the polymorphic base
    for UserSettings and SystemSettings.
    """

    __tablename__ = "settings"
    key = db.Column("StKey", db.String(40), primary_key=True)
    keytype = db.Column("StKeyType", db.String(10), nullable=False)
    value = db.Column("StValue", db.String, nullable=False)
    __mapper_args__ = {"polymorphic_on": keytype}

    @classmethod
    def key_exists_precheck(cls, keyname):
        """
        Check key validity for certain actions.
        """

    @classmethod
    def set_value_post(cls, keyname, keyvalue):
        """
        Post-setting value for certain keys."
        """

    @classmethod
    def set_value(cls, keyname, keyvalue):
        "Set, but don't save, a setting."
        cls.key_exists_precheck(keyname)
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is None:
            s = cls()
            s.key = keyname
        s.value = keyvalue
        db.session.add(s)
        cls.set_value_post(keyname, keyvalue)

    @classmethod
    def key_exists(cls, keyname):
        "True if exists."
        s = db.session.query(cls).filter(cls.key == keyname).first()
        no_key = s is None
        return not no_key

    @classmethod
    def get_value(cls, keyname):
        "Get the saved key, or None if it doesn't exist."
        cls.key_exists_precheck(keyname)
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is None:
            return None
        return s.value

    @classmethod
    def delete_key(cls, keyname):
        "Delete a key."
        s = db.session.query(cls).filter(cls.key == keyname).first()
        if s is not None:
            db.session.delete(s)


class MissingUserSettingKeyException(Exception):
    """
    Cannot set or get unknown user keys.
    """


class UserSetting(SettingBase):
    "User setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "user"}

    @classmethod
    def key_exists_precheck(cls, keyname):
        """
        User keys must exist.
        """
        if not UserSetting.key_exists(keyname):
            raise MissingUserSettingKeyException(keyname)

    @classmethod
    def set_value_post(cls, keyname, keyvalue):
        """
        Setting some keys runs other code.
        """
        if keyname == "mecab_path":
            mp = "MECAB_PATH"
            if keyvalue is None or keyvalue == "":
                if mp in os.environ:
                    del os.environ[mp]
            else:
                os.environ[mp] = keyvalue.strip()

    @staticmethod
    def _revised_mecab_path():
        """
        Change the mecab_path if it's not found, and a
        replacement is found.

        Lute Docker images are built to be multi-arch, and
        interestingly (annoyingly), mecab libraries are installed into
        different locations depending on the architecture, even with
        the same Dockerfile and base image.

        Returns: new mecab path if old one is missing _and_
        new one found, otherwise just return the old one.
        """

        mp = UserSetting.get_value("mecab_path")
        if mp is not None and os.path.exists(mp):
            return mp

        # See develop docs for notes on how to find the libmecab path!
        candidates = [
            # linux/arm64
            "/lib/aarch64-linux-gnu/libmecab.so.2",
            # linux/amd64
            "/lib/x86_64-linux-gnu/libmecab.so.2",
            # github CI, ubuntu-latest
            "/lib/x86_64-linux-gnu/libmecab.so.2",
        ]
        replacements = [p for p in candidates if os.path.exists(p)]
        if len(replacements) > 0:
            return replacements[0]
        # Replacement not found, leave current value as-is.
        return mp

    @staticmethod
    def load():
        """
        Load missing user settings with default values.
        """
        app_config = current_app.env_config

        # These keys are rendered into the global javascript namespace var
        # LUTE_USER_SETTINGS, so if any of these keys change, check the usage
        # of that variable as well.
        keys_and_defaults = {
            "backup_enabled": True,
            "backup_auto": True,
            "backup_warn": True,
            "backup_dir": app_config.default_user_backup_path,
            "backup_count": 5,
            "mecab_path": None,
            "japanese_reading": "katakana",
            "current_theme": "-",
            "custom_styles": "/* Custom css to modify Lute's appearance. */",
            "show_highlights": True,
            "current_language_id": 0,
            # Behaviour:
            "open_popup_in_new_tab": False,
            "stop_audio_on_term_form_open": True,
        }
        for k, v in keys_and_defaults.items():
            if not UserSetting.key_exists(k):
                s = UserSetting()
                s.key = k
                s.value = v
                db.session.add(s)
        db.session.commit()

        # Revise the mecab path if necessary.
        # Note this is done _after_ the defaults are loaded,
        # because the user may have already loaded the defaults
        # (e.g. on machine upgrade) and stored them in the db,
        # so we may have to _update_ the existing setting.
        revised_mecab_path = UserSetting._revised_mecab_path()
        UserSetting.set_value("mecab_path", revised_mecab_path)
        db.session.commit()

    @staticmethod
    def all_settings():
        """
        Get dict of all settings, for rendering into Javascript global space.
        """
        settings = db.session.query(UserSetting).all()
        ret = {}
        for s in settings:
            ret[s.key] = s.value

        # Convert some ints into bools.
        boolkeys = ["open_popup_in_new_tab", "stop_audio_on_term_form_open"]
        for k in boolkeys:
            ret[k] = ret[k] == "1"

        return ret


class SystemSetting(SettingBase):
    "System setting."
    __tablename__ = None
    __mapper_args__ = {"polymorphic_identity": "system"}

    # Helpers for certain sys settings.

    @classmethod
    def get_last_backup_datetime(cls):
        "Get the last_backup_datetime as int, or None."
        v = cls.get_value("lastbackup")
        if v is None:
            return None
        return int(v)

    @classmethod
    def set_last_backup_datetime(cls, v):
        "Set and save the last backup time."
        cls.set_value("lastbackup", v)
        db.session.commit()


class BackupSettings:
    """
    Convenience wrapper for current backup settings.
    Getter only.
    """

    def __init__(self):
        def _bool(k):
            v = UserSetting.get_value(k)
            return v in (1, "1", "y", True)

        self.backup_enabled = _bool("backup_enabled")
        self.backup_auto = _bool("backup_auto")
        self.backup_warn = _bool("backup_warn")
        self.backup_dir = UserSetting.get_value("backup_dir")
        self.backup_count = int(UserSetting.get_value("backup_count") or 5)
        self.last_backup_datetime = SystemSetting.get_last_backup_datetime()

    @property
    def last_backup_display_date(self):
        "Return the last_backup_datetime as yyyy-mm etc., or None if not set."
        t = self.last_backup_datetime
        if t is None:
            return None
        return datetime.datetime.fromtimestamp(t).strftime("%Y-%m-%d %H:%M:%S")

    @staticmethod
    def get_backup_settings():
        "Get BackupSettings."
        return BackupSettings()

    @property
    def time_since_last_backup(self):
        """
        Return the time since the last backup. Returns None either if not set or
        it is in the future.
        Eg. "3 days ago"
        """
        t = self.last_backup_datetime
        if t is None:
            return None

        delta = int(time.time() - t)
        if delta < 0:
            return None

        thresholds = [
            ("week", 1 * 60 * 60 * 24 * 7),
            ("day", 1 * 60 * 60 * 24),
            ("hour", 1 * 60 * 60),
            ("minute", 1 * 60),
            ("second", 1),
        ]

        for unit, seconds in thresholds:
            multiples = abs(delta // seconds)
            if multiples >= 1:
                message = f"{multiples} {unit}"
                if multiples > 1:
                    message += "s"
                break
        else:
            message = f"{abs(delta)} seconds"

        return message + " ago"



"""
Term parent mapping.
"""

from lute.db import db
from lute.models.term import Term


## Exports


# TODO issue_336_export_unknown_book_terms: move this where needed.
def export_unknown_terms(book, outfile):
    "Export unknown terms in the book to outfile."
    lang = book.language
    unique_tokens = {
        t for txt in book.texts for t in lang.get_parsed_tokens(txt.text) if t.is_word
    }
    unique_lcase_toks = {lang.get_lowercase(t.token) for t in unique_tokens}

    lgid = lang.id
    known_terms_lc = (
        db.session.query(Term.text_lc)
        .filter(Term.language_id == lgid, Term.token_count == 1)
        .all()
    )
    known_terms_lc = [word[0] for word in known_terms_lc]

    newtoks = [t for t in unique_lcase_toks if t not in known_terms_lc]
    with open(outfile, "w", encoding="utf-8") as f:
        f.write("\n".join(newtoks))



"""
Parser registry tests.
"""

import pytest

from lute.parse.registry import (
    __LUTE_PARSERS__,
    get_parser,
    supported_parsers,
    is_supported,
)
from lute.parse.space_delimited_parser import SpaceDelimitedParser


def test_get_parser_by_name():
    p = get_parser("spacedel")
    assert isinstance(p, SpaceDelimitedParser)


def test_get_parser_throws_if_not_found():
    with pytest.raises(ValueError):
        get_parser("trash")


def test_supported_parsers():
    "Sanity check only."
    d = supported_parsers()
    assert isinstance(d, list), "returns a list"

    p = [n for n in d if n[0] == "spacedel"][0]
    assert p == ("spacedel", "Space Delimited"), "sanity check"


class DummyParser:
    "Dummy unsupported parser."

    @classmethod
    def is_supported(cls):
        return False

    @classmethod
    def name(cls):
        return "DUMMY"


@pytest.fixture(name="_load_dummy")
def fixture_load_dummy():
    "Add the dummy parser for the test."
    __LUTE_PARSERS__["dummy"] = DummyParser
    yield
    del __LUTE_PARSERS__["dummy"]


def test_unavailable_parser_not_included_in_lists(_load_dummy):
    "An unsupported parser shouldn't be available."
    d = supported_parsers()
    assert "dummy" not in d, "not present"
    assert is_supported("dummy") is False, "no"
    with pytest.raises(ValueError):
        get_parser("dummy")



"""
DB migration tests.
"""

import sqlite3
from contextlib import closing
import pytest

from lute.db.setup.migrator import SqliteMigrator


@pytest.fixture(name="db")
def fixture_db(tmp_path):
    """
    Create a db for migrations, return the connection.
    """
    dbfile = tmp_path / "test.db"

    sql = """
    create table A (a int);

    CREATE TABLE IF NOT EXISTS "_migrations" (
	"filename" VARCHAR(255) NOT NULL  ,
	PRIMARY KEY ("filename")
    );
    """
    with closing(sqlite3.connect(dbfile)) as conn:
        conn.executescript(sql)
        yield conn


@pytest.fixture(name="migdirs")
def fixture_migdirs(tmp_path):
    """
    Create folders for migration files.
    """
    migs = tmp_path / "migrations"
    migs.mkdir()
    reps = tmp_path / "repeatable"
    reps.mkdir()
    return tmp_path


def test_no_migrations(db, migdirs):
    """
    No migrations if no unapplied files yet.
    """
    migrator = SqliteMigrator(migdirs / "migrations", migdirs / "repeatable")
    assert migrator.has_migrations(db) is False, "no migrations"

    migfile = migdirs / "migrations" / "a.sql"
    migfile.write_text("create table B (b int);")
    assert migrator.has_migrations(db) is True, "have migrations"


def test_migration_applied(db, migdirs):
    """
    Migrations are applied, and tracked.
    """
    migrator = SqliteMigrator(migdirs / "migrations", migdirs / "repeatable")

    migfile = migdirs / "migrations" / "b.sql"
    migfile.write_text("create table B (b int);")
    assert migrator.has_migrations(db) is True, "have migrations"

    migrator.do_migration(db)

    assert migrator.has_migrations(db) is False, "no outstanding migrations"

    cur = db.cursor()
    sql = """SELECT name FROM sqlite_master
    WHERE type='table' AND name in ('A', 'B')
    order by name;"""
    tnames = cur.execute(sql).fetchall()
    tnames = [t[0] for t in tnames]
    assert tnames == ["A", "B"], "migrations run"


def test_bad_migration_throws(db, migdirs):
    """
    Bad migration throws.
    """
    migrator = SqliteMigrator(migdirs / "migrations", migdirs / "repeatable")

    migfile = migdirs / "migrations" / "b.sql"
    migfile.write_text("create tableXXXX B (b int);")
    assert migrator.has_migrations(db) is True, "have migrations"

    with pytest.raises(sqlite3.OperationalError):
        migrator.do_migration(db)

    assert migrator.has_migrations(db), "bad file still pending!!!"


def test_repeatable_migs_are_always_applied(db, migdirs):
    """
    Always applied, even if no migrations.
    """
    migrator = SqliteMigrator(migdirs / "migrations", migdirs / "repeatable")

    migfile = migdirs / "repeatable" / "r.sql"
    migfile.write_text("drop view if exists v1; create view v1 as select 1;")
    assert migrator.has_migrations(db) is False, "have migrations"
    migrator.do_migration(db)

    cur = db.cursor()
    vals = cur.execute("SELECT * from v1;").fetchall()
    assert vals == [(1,)], "migrations run"

    migfile = migdirs / "repeatable" / "r.sql"
    migfile.write_text("drop view if exists v1; create view v1 as select 2;")
    assert migrator.has_migrations(db) is False, "have migrations"
    migrator.do_migration(db)

    cur = db.cursor()
    vals = cur.execute("SELECT * from v1;").fetchall()
    assert vals == [(2,)], "changed"


def test_bad_repeatable_migs_throws(db, migdirs):
    """
    Bad repeatable should throw.
    """
    migrator = SqliteMigrator(migdirs / "migrations", migdirs / "repeatable")
    migfile = migdirs / "repeatable" / "r.sql"
    migfile.write_text("drop view if exists v1; create blahblah")

    with pytest.raises(sqlite3.OperationalError):
        migrator.do_migration(db)



"""
Parser registry.

List of available parsers.
"""

from importlib.metadata import entry_points
from sys import version_info

from lute.parse.base import AbstractParser
from lute.parse.space_delimited_parser import SpaceDelimitedParser, TurkishParser
from lute.parse.mecab_parser import JapaneseParser
from lute.parse.character_parser import ClassicalChineseParser


__LUTE_PARSERS__ = {
    "spacedel": SpaceDelimitedParser,
    "turkish": TurkishParser,
    "japanese": JapaneseParser,
    "classicalchinese": ClassicalChineseParser,
}


def init_parser_plugins():
    """
    Initialize parsers from plugins
    """

    vmaj = version_info.major
    vmin = version_info.minor
    if vmaj == 3 and vmin in (8, 9, 10, 11):
        custom_parser_eps = entry_points().get("lute.plugin.parse")
    elif (vmaj == 3 and vmin >= 12) or (vmaj >= 4):
        # Can't be sure this will always work, API may change again,
        # but can't plan for the unforseeable everywhere.
        custom_parser_eps = entry_points().select(group="lute.plugin.parse")
    else:
        # earlier version of python than 3.8?  What madness is this?
        # Not going to throw, just print and hope the user sees it.
        msg = f"Unable to load plugins for python {vmaj}.{vmin}, please upgrade to 3.8+"
        print(msg, flush=True)
        return

    if custom_parser_eps is None:
        return

    for custom_parser_ep in custom_parser_eps:
        if _is_valid(custom_parser_ep.load()):
            __LUTE_PARSERS__[custom_parser_ep.name] = custom_parser_ep.load()
        else:
            raise ValueError(
                f"{custom_parser_ep.name} is not a subclass of AbstractParser"
            )


def _is_valid(custom_parser):
    return issubclass(custom_parser, AbstractParser)


def _supported_parsers():
    "Get the supported parsers."
    ret = {}
    for k, v in __LUTE_PARSERS__.items():
        if v.is_supported():
            ret[k] = v
    return ret


def get_parser(parser_name) -> AbstractParser:
    "Return the supported parser with the given name."
    if parser_name in _supported_parsers():
        pclass = __LUTE_PARSERS__[parser_name]
        return pclass()
    raise ValueError(f"Unknown parser type '{parser_name}'")


def is_supported(parser_name) -> bool:
    "Return True if the specified parser is present and supported."
    if parser_name not in __LUTE_PARSERS__:
        return False
    p = __LUTE_PARSERS__[parser_name]
    return p.is_supported()


def supported_parsers():
    """
    List of supported parser strings and class names, for UI.

    For select list entries, use supported_parsers().items().
    """
    return [(k, v.name()) for k, v in _supported_parsers().items()]


def supported_parser_types():
    """
    List of supported Language.parser_types
    """
    return list(_supported_parsers().keys())



"""
Theming service.

Themes are stored in the css folder.  Current theme is saved in
UserSetting.
"""

import os
from glob import glob
from flask import current_app
from lute.models.setting import UserSetting
from lute.db import db

default_entry = ("-", "(default)")


def _css_path():
    """
    Path to css in this folder.
    """
    thisdir = os.path.dirname(__file__)
    theme_dir = os.path.join(thisdir, "css")
    return os.path.abspath(theme_dir)


def list_themes():
    """
    List of theme file names and user-readable name.
    """

    def _make_display_name(s):
        ret = os.path.basename(s)
        ret = ret.replace(".css", "").replace("_", " ")
        return ret

    g = glob(os.path.join(_css_path(), "*.css"))
    themes = [(os.path.basename(f), _make_display_name(f)) for f in g]
    theme_basenames = [t[0] for t in themes]

    g = glob(os.path.join(current_app.env_config.userthemespath, "*.css"))
    additional_user_themes = [
        (os.path.basename(f), _make_display_name(f))
        for f in g
        if os.path.basename(f) not in theme_basenames
    ]

    themes += additional_user_themes
    sorted_themes = sorted(themes, key=lambda x: x[1])
    return [default_entry] + sorted_themes


def get_current_css():
    """
    Return the current css pointed at by the current_theme user setting.
    """
    current_theme = UserSetting.get_value("current_theme")
    if current_theme == default_entry[0]:
        return ""

    def _get_theme_css_in_dir(d):
        "Get css, or '' if no file."
        fname = os.path.join(d, current_theme)
        if not os.path.exists(fname):
            return ""
        with open(fname, "r", encoding="utf-8") as f:
            return f.read()

    ret = _get_theme_css_in_dir(_css_path())
    add = _get_theme_css_in_dir(current_app.env_config.userthemespath)
    if add != "":
        ret += f"\n\n/* Additional user css */\n\n{add}"
    return ret


def next_theme():
    """
    Move to the next theme in the list of themes.
    """
    current_theme = UserSetting.get_value("current_theme")
    themes = [t[0] for t in list_themes()]
    themes.append(default_entry[0])
    for i in range(0, len(themes)):  # pylint: disable=consider-using-enumerate
        if themes[i] == current_theme:
            new_index = i + 1
            break
    UserSetting.set_value("current_theme", themes[new_index])
    db.session.commit()



"""
Book create/edit forms.
"""

import json
from flask import request
from wtforms import StringField, SelectField, TextAreaField, IntegerField, HiddenField
from wtforms import ValidationError
from wtforms.validators import DataRequired, Length, NumberRange
from flask_wtf import FlaskForm
from flask_wtf.file import FileField, FileAllowed
from lute.book import service


class NewBookForm(FlaskForm):
    """
    New book.  All fields can be entered.
    """

    language_id = SelectField("Language", coerce=int)

    title = StringField("Title", validators=[DataRequired(), Length(max=255)])

    desc = (
        "Use for short texts, e.g. up to a few thousand words. "
        + 'For longer texts, use the "Text file" below.'
    )
    text = TextAreaField("Text", description=desc)
    textfile = FileField(
        "Text file",
        validators=[
            FileAllowed(
                ["txt", "epub", "pdf", "srt", "vtt"],
                "Please upload a valid '.txt', '.epub', '.pdf', '.srt' or '.vtt' file.",
            )
        ],
    )
    max_page_tokens = IntegerField(
        "Words per page",
        validators=[NumberRange(min=10, max=1500)],
        default=250,
    )
    source_uri = StringField("Text source", validators=[Length(max=255)])
    audiofile = FileField(
        "Audio file",
        validators=[
            FileAllowed(
                ["mp3", "wav", "ogg", "opus"],
                "Please upload a valid audio file (mp3, wav, ogg, opus)",
            )
        ],
    )
    book_tags = StringField("Tags")

    def __init__(self, *args, **kwargs):
        "Call the constructor of the superclass (FlaskForm)"
        super().__init__(*args, **kwargs)
        book = kwargs.get("obj")

        def _data(arr):
            "Get data in proper format for tagify."
            return json.dumps([{"value": p} for p in arr])

        self.book_tags.data = _data(book.book_tags)
        if request.method == "POST":
            self.book_tags.data = request.form.get("book_tags", "")

    def populate_obj(self, obj):
        "Call the populate_obj method from the parent class, then mine."
        super().populate_obj(obj)

        def _values(field_data):
            "Convert field data to array."
            ret = []
            if field_data:
                ret = [h["value"] for h in json.loads(field_data)]
            return ret

        obj.book_tags = _values(self.book_tags.data)

        if self.textfile.data:
            obj.text = service.get_file_content(self.textfile.data)
        f = self.audiofile.data
        if f:
            obj.audio_filename = service.save_audio_file(f)

    def validate_language_id(self, field):  # pylint: disable=unused-argument
        "Language must be set."
        if self.language_id.data in (None, 0):
            raise ValidationError("Please select a language")

    def validate_text(self, field):  # pylint: disable=unused-argument
        "Throw if missing text and textfile, or if have both."
        have_text = self.text.data not in ("", None)
        have_textfile = self.textfile.data not in ("", None)
        if have_text and have_textfile:
            raise ValidationError(
                "Both Text and Text file are set, please only specify one"
            )
        if have_text is False and have_textfile is False:
            raise ValidationError("Please specify either Text or Text file")


class EditBookForm(FlaskForm):
    """
    Edit existing book.  Only a few fields can be changed.
    """

    title = StringField("Title", validators=[DataRequired(), Length(max=255)])
    source_uri = StringField("Source URI", validators=[Length(max=255)])
    book_tags = StringField("Tags")
    audiofile = FileField(
        "Audio file",
        validators=[
            FileAllowed(
                ["mp3", "wav", "ogg", "opus"],
                "Please upload a valid audio file (mp3, wav, ogg, opus)",
            )
        ],
    )

    # The current audio_filename can be removed from the current book.
    audio_filename = HiddenField("Audio filename")

    def __init__(self, *args, **kwargs):
        "Call the constructor of the superclass (FlaskForm)"
        super().__init__(*args, **kwargs)
        book = kwargs.get("obj")

        def _data(arr):
            "Get data in proper format for tagify."
            return json.dumps([{"value": p} for p in arr])

        self.book_tags.data = _data(book.book_tags)
        if request.method == "POST":
            self.book_tags.data = request.form.get("book_tags", "")

    def populate_obj(self, obj):
        "Call the populate_obj method from the parent class, then mine."
        super().populate_obj(obj)

        def _values(field_data):
            "Convert field data to array."
            ret = []
            if field_data:
                ret = [h["value"] for h in json.loads(field_data)]
            return ret

        obj.book_tags = _values(self.book_tags.data)

        f = self.audiofile.data
        if f:
            obj.audio_filename = service.save_audio_file(f)
            obj.audio_bookmarks = None
            obj.audio_current_pos = None



"""
TermForm tests.

Tests lute.term.model.Term *domain* objects being saved
and retrieved from DB.
"""

from lute.models.term import Term as DBTerm
from lute.db import db
from lute.term.model import Term, Repository
from lute.term.forms import TermForm


def test_validate(app, app_context, english):
    "A new term is valid."
    repo = Repository(db)
    t = repo.find_or_new(english.id, "CAT")
    with app.test_request_context():
        f = TermForm(obj=t)
    f.language_id.choices = [(english.id, "english")]

    assert f.validate() is True, "no change = valid"


def test_text_change_not_valid(app, app_context, english):
    "Text change raises error."
    dbt = DBTerm(english, "CAT")
    db.session.add(dbt)
    db.session.commit()

    repo = Repository(db)
    t = repo.find_or_new(english.id, "CAT")
    t.text = "dog"
    with app.test_request_context():
        f = TermForm(obj=t)
    f.language_id.choices = [(english.id, "english")]

    is_valid = f.validate()
    assert is_valid is False, "text change = not valid"
    assert f.errors == {"text": ["Can only change term case"]}


def test_duplicate_text_not_valid(app, app_context, english):
    "Duplicate term throws."
    dbt = DBTerm(english, "CAT")
    db.session.add(dbt)
    db.session.commit()

    t = Term()
    t.language_id = english.id
    t.text = "cat"
    with app.test_request_context():
        f = TermForm(obj=t)
    f.language_id.choices = [(english.id, "english")]

    is_valid = f.validate()
    assert is_valid is False, "dup term not valid"
    assert f.errors == {"text": ["Term already exists"]}


def test_update_existing_term_is_valid(app, app_context, english):
    "Can update an existing term."
    dbt = DBTerm(english, "CAT")
    db.session.add(dbt)
    db.session.commit()

    repo = Repository(db)
    t = repo.find_or_new(english.id, "cat")
    t.text = "cat"
    with app.test_request_context():
        f = TermForm(obj=t)
    f.language_id.choices = [(english.id, "english")]

    is_valid = f.validate()
    assert is_valid is True, "updating existing term is ok"



"""
Simple CLI commands.
"""

import click
from flask import Blueprint

from lute.cli.language_term_export import generate_file

bp = Blueprint("cli", __name__)


@bp.cli.command("hello")
def hello():
    "Say hello -- proof-of-concept CLI command only."
    msg = """
    Hello there!

    This is the Lute cli.

    There may be some experimental scripts here ...
    nothing that will change or damage your Lute data,
    but the CLI may change.

    Thanks for looking.
    """
    print(msg)


@bp.cli.command("language_export")
@click.argument("language")
@click.argument("output_path")
def language_export(language, output_path):
    """
    Get all terms from active books in the language, and write a
    data file of term frequencies and children.
    """
    generate_file(language, output_path)



"""
Common form methods.
"""

from lute.models.language import Language
from lute.models.setting import UserSetting
from lute.db import db


def language_choices(dummy_entry_placeholder="-"):
    """
    Return the list of languages for select boxes.

    If only one lang exists, only return that,
    otherwise add a '-' dummy entry at the top.
    """
    langs = db.session.query(Language).order_by(Language.name).all()
    supported = [lang for lang in langs if lang.is_supported]
    lang_choices = [(s.id, s.name) for s in supported]
    # Add a dummy placeholder even if there are no languages.
    if len(lang_choices) != 1:
        lang_choices = [(0, dummy_entry_placeholder)] + lang_choices
    return lang_choices


def valid_current_language_id():
    """
    Get the current language id from UserSetting, ensuring
    it's still valid.  If not, change it.
    """
    current_language_id = UserSetting.get_value("current_language_id")
    current_language_id = int(current_language_id)

    valid_language_ids = [int(p[0]) for p in language_choices()]
    if current_language_id in valid_language_ids:
        return current_language_id

    current_language_id = valid_language_ids[0]
    UserSetting.set_value("current_language_id", current_language_id)
    db.session.commit()
    return current_language_id
